package core

import (
	"git.akyoto.dev/cli/q/src/build/asm"
	"git.akyoto.dev/cli/q/src/build/ast"
	"git.akyoto.dev/cli/q/src/build/cpu"
	"git.akyoto.dev/cli/q/src/build/errors"
	"git.akyoto.dev/cli/q/src/build/expression"
	"git.akyoto.dev/cli/q/src/build/token"
	"git.akyoto.dev/cli/q/src/build/types"
)

// ExpressionToRegister puts the result of an expression into the specified register.
func (f *Function) ExpressionToRegister(node *expression.Expression, register cpu.Register) (types.Type, error) {
	f.SaveRegister(register)

	if node.IsFolded {
		f.RegisterNumber(asm.MOVE, register, node.Value)
		return types.Int, nil
	}

	if node.IsLeaf() {
		return f.TokenToRegister(node.Token, register)
	}

	if ast.IsFunctionCall(node) {
		fn, err := f.CompileCall(node)

		if register != f.CPU.Output[0] {
			f.RegisterRegister(asm.MOVE, register, f.CPU.Output[0])
		}

		if fn == nil || len(fn.ReturnTypes) == 0 {
			return types.Invalid, err
		}

		return fn.ReturnTypes[0], err
	}

	if node.Token.Kind == token.Array {
		array := f.VariableByName(node.Children[0].Token.Text(f.File.Bytes))
		offset, err := f.Number(node.Children[1].Token)
		f.MemoryRegister(asm.LOAD, asm.Memory{Base: array.Register, Offset: byte(offset), Length: 1}, register)
		return types.Int, err
	}

	if len(node.Children) == 1 {
		if !node.Token.IsUnaryOperator() {
			return types.Invalid, errors.New(errors.MissingOperand, f.File, node.Token.End())
		}

		typ, err := f.ExpressionToRegister(node.Children[0], register)

		if err != nil {
			return typ, err
		}

		return typ, f.ExecuteRegister(node.Token, register)
	}

	left := node.Children[0]
	right := node.Children[1]
	final := register

	if f.UsesRegister(right, register) {
		register = f.NewRegister()
	}

	typ, err := f.ExpressionToRegister(left, register)

	if err != nil {
		return types.Invalid, err
	}

	err = f.Execute(node.Token, register, right)

	if register != final {
		f.RegisterRegister(asm.MOVE, final, register)
		f.FreeRegister(register)
	}

	return typ, err
}
